from uuid import uuid4
from langchain_core.messages import ToolMessage, HumanMessage
from rich import print
from shy_sh.agents.tools.shell_exe import shell_exec
from shy_sh.models import State
from shy_sh.agents.tools import tools_by_name
from shy_sh.settings import settings
from shy_sh.agents.misc import parse_react_tool


def tools_handler(state: State):
    last_message = state["tool_history"][-1]

    tool_answers = []
    executed_scripts = []
    if settings.llm.agent_pattern == "flow":
        # check message scripts
        if "```" in last_message.content:
            script = last_message.content.split("```")[1].strip()
            script_type = script.split("\n", 1)[0].strip().split(" ")[0].strip()
            script = script.replace(script_type, "", 1).strip()
            if not script:
                return {}
            try:
                print()
                result, art = shell_exec(
                    state,
                    script,
                    script_type,
                    ask_before_execute=state["ask_before_execute"],
                )
                message = HumanMessage(content=result)
                message.artifact = art
                executed_scripts += [
                    {
                        "script": script,
                        "type": script_type,
                        "result": message.content,
                    }
                ]
            except Exception as e:
                print(f"[bold red]🚨 Execution error: {e}[/bold red]")
                message = HumanMessage(f"Execution error: {e}")
            tool_answers.append(message)
        return {"tool_history": tool_answers, "executed_scripts": executed_scripts}

    t_calls = _get_tool_calls(last_message)
    for t_call in t_calls:
        try:
            t = tools_by_name[t_call["name"]]
            message = t.invoke(
                {
                    **t_call,
                    "args": {"state": state, **t_call["args"]},
                }
            )
        except Exception as e:
            print(f"[bold red]🚨 Tool error: {e}[/bold red]")
            message = ToolMessage(f"Tool error: {e}", tool_call_id=t_call["id"])

        if settings.llm.agent_pattern == "react":
            m = HumanMessage(content=f"Tool response:\n{message.content}")
            m.artifact = getattr(message, "artifact", None)
            message = m
        tool_answers.append(message)
        if hasattr(message, "artifact") and message.artifact.executed_scripts:
            executed_scripts += message.artifact.executed_scripts
    return {"tool_history": tool_answers, "executed_scripts": executed_scripts}


def _get_react_tool_calls(message):
    react_tool = parse_react_tool(message)
    return [
        {
            "name": react_tool.tool,
            "args": {"arg": react_tool.arg},
            "id": uuid4().hex,
            "type": "tool_call",
        }
    ]


def _get_function_call_tool_calls(message):
    return message.tool_calls


def _get_tool_calls(message):
    match (settings.llm.agent_pattern):
        case "react":
            return _get_react_tool_calls(message)
        case "function_call":
            return _get_function_call_tool_calls(message)
        case _:
            return []



import pyperclip
from typing import Annotated
from rich import print
from langgraph.prebuilt import InjectedState
from langchain.tools import tool
from questionary import select, Style, Choice
from shy_sh.models import State, ToolMeta
from shy_sh.utils import (
    ask_confirm,
    run_command,
    tools_to_human,
    detect_shell,
    detect_os,
)
from shy_sh.agents.chains.explain import explain
from shy_sh.agents.chains.alternative_commands import get_alternative_commands
from shy_sh.settings import settings

_text_style = {
    "qmark": "",
    "style": Style.from_dict(
        {
            "selected": "fg:ansigreen noreverse bold",
            "question": "fg:darkorange nobold",
            "highlighted": "fg:ansigreen bold",
            "text": "fg:ansigreen bold",
            "answer": "fg:ansigreen bold",
            "instruction": "fg:ansigreen",
        }
    ),
}

_select_style = {
    "pointer": "►",
    "instruction": " ",
    **_text_style,
}


@tool(response_format="content_and_artifact")
def shell(arg: str, state: Annotated[State, InjectedState]):
    """to execute a shell command in the terminal, useful for every task that requires to interact with the current system or local files, do not pass multiple lines commands, avoid to install new packages if not explicitly requested"""
    print(f"🛠️ [bold green] {arg} [/bold green]")
    result = ""
    confirm = "y"
    if state["ask_before_execute"]:
        confirm = ask_confirm(alternatives=True)
    print()
    if confirm == "n":
        return "Command interrupted by the user", ToolMeta(
            stop_execution=True, skip_print=True
        )
    elif confirm == "c":
        pyperclip.copy(arg)
        return "Command copied to the clipboard!", ToolMeta(stop_execution=True)
    elif confirm == "a":
        r = _select_alternative_command(arg, state)
        print()
        if r == "None":
            return "Command interrupted by the user", ToolMeta(
                stop_execution=True, skip_print=True
            )
        if settings.sandbox_mode:
            pyperclip.copy(r)
            return "Command copied to the clipboard!", ToolMeta(stop_execution=True)
        arg = r
        result += f"The user decided to execute this alternative command `{arg}`\n\n"
    elif confirm == "e":
        inputs = {
            "task": state["history"][-1].content,
            "script_type": "shell command",
            "script": arg,
            "timestamp": state["timestamp"],
        }
        ret = explain(inputs, ask_alternative=True)
        if ret == "alternative":
            r = _select_alternative_command(arg, state)
            print()
            if r == "None":
                return "Command interrupted by the user", ToolMeta(
                    stop_execution=True, skip_print=True
                )
            arg = r
            result += (
                f"The user decided to execute this alternative command `{arg}`\n\n"
            )
        elif ret:
            return ret
    output = run_command(arg)
    result += output

    if len(result) > 20000:
        print("\n🐳 [bold red]Output too long! It will be truncated[/bold red]")
        result = (
            result[:9000] + "\n...(OUTPUT TOO LONG TRUNCATED!)...\n" + result[-9000:]
        )
    return result, ToolMeta(
        executed_scripts=[{"script": arg, "type": "shell", "result": output}]
    )


def _select_alternative_command(arg, state):
    inputs = {
        "timestamp": state["timestamp"],
        "shell": detect_shell(),
        "system": detect_os(),
        "history": tools_to_human(state["history"] + state["tool_history"]),
        "cmd": arg,
    }
    cmds = get_alternative_commands(inputs)
    r = select(
        (
            "Pick the command to copy to the clipboard"
            if settings.sandbox_mode
            else "Pick the command to execute"
        ),
        choices=[
            Choice([("fg:ansired bold", "Cancel")], "None"),
            Choice(
                [
                    ("fg:ansiyellow bold", arg),
                    ("fg:gray", " # Original command"),
                ],
                arg,
            ),
            *[
                Choice(
                    [("fg:ansigreen bold", c[1]), ("fg:gray", " " + c[0])],
                    c[1],
                )
                for c in cmds
            ],
        ],
        **_select_style,
    ).unsafe_ask()

    return r



import os
import sys
import pyperclip
from typing import Annotated
from tempfile import NamedTemporaryFile
from rich import print
from rich.live import Live
from langgraph.prebuilt import InjectedState
from langchain.tools import tool
from shy_sh.models import State, ToolMeta
from shy_sh.utils import ask_confirm, detect_shell, detect_os, parse_code
from shy_sh.agents.chains.shell_expert import shexpert_chain
from shy_sh.utils import run_command, tools_to_human, syntax
from shy_sh.agents.chains.explain import explain


@tool(response_format="content_and_artifact")
def shell_expert(arg: str, state: Annotated[State, InjectedState]):
    """to delegate the task to a shell expert that can write and execute long and complex shell scripts, use only if you cant resolve the task with a simple shell command, just explain what do you want to achieve in a short sentence in the arg without including any shell code"""
    print(f"💻 [bold dark_green]Generating shell script...[/bold dark_green]\n")
    shell = detect_shell()
    system = detect_os()
    inputs = {
        "input": arg,
        "system": system,
        "shell": shell,
        "timestamp": state["timestamp"],
        "history": tools_to_human(state["history"] + state["tool_history"]),
    }
    code = ""
    with Live(vertical_overflow="visible") as live:
        for chunk in shexpert_chain.stream(inputs):
            code += chunk  # type: ignore
            live.update(syntax(code))

        code = parse_code(code)
        live.update(syntax(code.strip()))

    confirm = "y"
    if state["ask_before_execute"]:
        confirm = ask_confirm()
    print()
    if confirm == "n":
        return "Script interrupted by the user", ToolMeta(
            stop_execution=True, skip_print=True
        )
    elif confirm == "c":
        pyperclip.copy(code)
        return "Script copied to the clipboard!", ToolMeta(stop_execution=True)
    elif confirm == "e":
        inputs = {
            "task": arg,
            "script_type": "shell script",
            "script": code,
            "timestamp": state["timestamp"],
        }
        ret = explain(inputs)
        if ret:
            return ret

    ext = ".sh"
    if shell == "cmd":
        ext = ".bat"
    elif shell == "powershell":
        ext = ".ps1"

    if sys.version_info >= (3, 12):
        with NamedTemporaryFile("w+", suffix=ext, delete_on_close=False) as file:
            file.write(code)
            file.close()
            os.chmod(file.name, 0o755)
            result = run_command(file.name)

            if len(result) > 20000:
                print("\n🐳 [bold red]Output too long! It will be truncated[/bold red]")
                result = (
                    result[:9000]
                    + "\n...(OUTPUT TOO LONG TRUNCATED!)...\n"
                    + result[-9000:]
                )

    else:
        with NamedTemporaryFile("w+", suffix=ext, delete=False) as file:
            file.write(code)
            file.close()
            os.chmod(file.name, 0o755)
            result = run_command(file.name)

            if len(result) > 20000:
                print("\n🐳 [bold red]Output too long! It will be truncated[/bold red]")
                result = (
                    result[:9000]
                    + "\n...(OUTPUT TOO LONG TRUNCATED!)...\n"
                    + result[-9000:]
                )
            os.unlink(file.name)
    print()
    ret = f"Script executed:\n{code}\n\nOutput:\n{result}"
    if len(ret) > 20000:
        ret = f"Output:\n{result}"
    return ret, ToolMeta(
        executed_scripts=[{"script": code, "type": "shell_script", "result": result}]
    )



from shy_sh.agents.misc import has_tool_calls, parse_react_tool
from shy_sh.manager.sql_models import ExecutedScript, ScriptType, session
from shy_sh.manager.chat_manager import ChatManager
from langchain_core.messages import HumanMessage, AIMessage, ToolMessage, BaseMessage
from rich import print
from rich.markdown import Markdown

from shy_sh.utils import syntax, to_local


def print_chat_history():
    print("\n[bold green]Chats[/bold green]")
    with session() as db:
        chat_manager = ChatManager(db)
        chats = chat_manager.get_all_chats()
        if not chats:
            print("No chat history found.")
            return

        for chat in chats:
            print(f"[bold magenta][{chat.id}][/] - [bold green]{chat.title}[/]")
            print(
                f"  Model: {chat.meta.get('provider', '')} - {chat.meta.get('name', 'Unknown')}"
            )
            print(f"  Messages: {len(chat.messages)}")
            print(f"  Created: {to_local(chat.created_at).strftime('%Y-%m-%d %H:%M')}")


_MESSAGE_MAP = {
    "human": HumanMessage,
    "ai": AIMessage,
    "tool": ToolMessage,
}


def _serialize_messages(messages: list[dict]) -> list[BaseMessage]:
    return [
        _MESSAGE_MAP[message["type"]].model_validate(message) for message in messages
    ]


def load_chat_history(chat_id: int) -> list[BaseMessage]:
    with session() as db:
        chat_manager = ChatManager(db)
        chat = chat_manager.get_chat(chat_id)
        if not chat:
            return []
        return _serialize_messages(chat.messages)


def print_chat(chat_id: int):
    hist = load_chat_history(chat_id)
    print(Markdown("---"))
    print()
    for message in hist:
        if hasattr(message, "tool_calls") and len(message.tool_calls) > 0:  # type: ignore
            print(syntax(message.tool_calls[0]["args"].get("arg")))  # type: ignore
            print()
        else:
            try:
                rtool = parse_react_tool(message)
            except Exception:
                rtool = None
            if rtool:
                print(syntax(rtool.arg))
            else:
                avatar = "✨" if isinstance(message, HumanMessage) else "🤖"
                msg = str(message.content)
                if msg.startswith("Tool response:"):
                    msg = msg[14:].strip()
                print(Markdown(f"{avatar}: {msg}"))
            print()
    print(Markdown("---"))


def print_recent_commands(
    script_type: ScriptType | None = None,
):
    print("\n[bold green]Commands history[/bold green]")
    with session() as db:
        chat_manager = ChatManager(db)
        scripts = chat_manager.get_recent_scripts(script_type)
        for script, count, kind, created_at in reversed(scripts):
            if len(script) > 100:
                script = script[:100] + "..."
            print(
                f"\n[magenta][{kind}][/] {to_local(created_at).strftime('%Y-%m-%d %H:%M')}"
            )
            print(
                syntax(
                    script, lexer="python" if kind == ScriptType.PYTHON else "console"
                )
            )


def save_chat_history(
    *,
    chat_id: int | None = None,
    title: str | None = None,
    messages: list[BaseMessage] | None = None,
    meta: dict | None = None,
    executed_scripts: list[dict] | None = None,
):
    with session() as db:
        chat_manager = ChatManager(db)
        chat_manager.save_chat(
            chat_id,
            title=title,
            messages=(
                [message.model_dump() for message in messages] if messages else None
            ),
            meta=meta,
            executed_scripts=(
                [ExecutedScript(**script) for script in executed_scripts]
                if executed_scripts
                else None
            ),
        )


def truncate_chats(keep: int = 100):
    with session() as db:
        chat_manager = ChatManager(db)
        chat_manager.truncate_chats(keep)



import os
import sys
from time import strftime
import pyperclip
from tempfile import NamedTemporaryFile
from questionary import Choice, Style, select
from rich import print
from shy_sh.agents.chains.alternative_commands import get_alternative_commands
from shy_sh.models import ToolMeta
from shy_sh.settings import settings
from shy_sh.utils import (
    ask_confirm,
    detect_os,
    detect_shell,
    run_python,
    tools_to_human,
)
from shy_sh.utils import run_command
from shy_sh.agents.chains.explain import explain

_text_style = {
    "qmark": "",
    "style": Style.from_dict(
        {
            "selected": "fg:ansigreen noreverse bold",
            "question": "fg:darkorange nobold",
            "highlighted": "fg:ansigreen bold",
            "text": "fg:ansigreen bold",
            "answer": "fg:ansigreen bold",
            "instruction": "fg:ansigreen",
        }
    ),
}

_select_style = {
    "pointer": "►",
    "instruction": " ",
    **_text_style,
}


def shell_exec(state, code: str, script_type: str, ask_before_execute: bool = True):
    if "\n" not in code:
        print(f"🛠️ [bold green] {code} [/bold green]")

    confirm = "y"
    result = ""
    ask_alternative = script_type != "python" and "\n" not in code
    if ask_before_execute:
        confirm = ask_confirm(alternatives=ask_alternative)
    print()
    if confirm == "n":
        return "Script not executed by the user", ToolMeta(
            stop_execution=True, skip_print=True
        )
    elif confirm == "a":
        r = _select_alternative_command(code, state)
        print()
        if r == "None":
            return shell_exec(state, code, script_type, ask_before_execute)
        code = r
        result = f"The user decided to execute this alternative command `{code}`\n\n"
    elif confirm == "c":
        pyperclip.copy(code)
        return "Script copied to the clipboard!", ToolMeta(stop_execution=True)
    elif confirm == "e":
        inputs = {
            "task": code,
            "script_type": script_type,
            "script": code,
            "timestamp": state["timestamp"],
        }
        explain(inputs, ask_execute=False)
        return shell_exec(state, code, script_type, ask_before_execute)

    if settings.sandbox_mode:
        pyperclip.copy(code)
        return "Script copied to the clipboard!", ToolMeta(stop_execution=True)
    ext = ".sh"
    shell = detect_shell()
    if script_type == "python":
        ext = ".py"
    elif shell == "cmd":
        ext = ".bat"
    elif shell == "powershell":
        ext = ".ps1"

    if sys.version_info >= (3, 12):
        with NamedTemporaryFile("w+", suffix=ext, delete_on_close=False) as file:
            file.write(code)
            file.close()
            os.chmod(file.name, 0o755)
            if script_type == "python":
                result += run_python(file.name)
            else:
                result += run_command(file.name)

    else:
        with NamedTemporaryFile("w+", suffix=ext, delete=False) as file:
            file.write(code)
            file.close()
            os.chmod(file.name, 0o755)
            if script_type == "python":
                result += run_python(file.name)
            else:
                result += run_command(file.name)
            os.unlink(file.name)

    if len(result) > 20000:
        print("\n🐳 [bold red]Output too long! It will be truncated[/bold red]")
        result = (
            result[:9000] + "\n...(OUTPUT TOO LONG TRUNCATED!)...\n" + result[-9000:]
        )
    print()
    return result, ToolMeta(
        executed_scripts=[{"script": code, "type": "shell_script", "result": result}]
    )


def _select_alternative_command(arg: str, state: dict):
    inputs = {
        "timestamp": state["timestamp"],
        "shell": detect_shell(),
        "system": detect_os(),
        "history": tools_to_human(state["history"] + state["tool_history"]),
        "cmd": arg,
    }
    cmds = get_alternative_commands(inputs)
    r = select(
        (
            "Pick the command to copy to the clipboard"
            if settings.sandbox_mode
            else "Pick the command to execute"
        ),
        choices=[
            Choice([("fg:ansired bold", "Cancel")], "None"),
            Choice(
                [
                    ("fg:ansiyellow bold", arg),
                    ("fg:gray", " # Original command"),
                ],
                arg,
            ),
            *[
                Choice(
                    [("fg:ansigreen bold", c[1]), ("fg:gray", " " + c[0])],
                    c[1],
                )
                for c in cmds
            ],
        ],
        **_select_style,
    ).unsafe_ask()

    return r



from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.runnables import chain
from shy_sh.agents.llms import get_llm
from shy_sh.settings import settings
from shy_sh.agents.tools import tools
from textwrap import dedent

SYS_TEMPLATES = {
    "flow": dedent(
        """
        You are a helpful shell assistant. The current date and time is {timestamp}.
        Solve the tasks that I request you to do.
        <rules>
        You can suggest only shell and python scripts. 
        Prefer to use shell over python scripts.
        To suggest a script wrap it in a code block with the language specified.
        Emit MAX ONE code block per message.
        If the user accepts your suggestion you will receive the output of the runned script, continue to use the output of the script to solve the task.
        Stop suggesting scripts when the task is completed, output the final answer WITHOUT ANY code block and ask the user if they need anything else.
        Avoid to delete files if not explicitly requested by the user.
        You can use markdown to format your responses.
        Is important to suggest NOT MORE THAN ONE code block per message and wait the user response before continuing, all the code blocks will be automatically executed by the system so NEVER use any kind of code blocks for the final answer.
        NEVER invent answers or pretend to know information that you do not have, instead you can suggest a script to gather the information you need.
        </rules>
        Answer truthfully with the informations you have. Output your answers in {lang_spec} language.
        """
    ),
    "react": dedent(
        """
        You are a helpful shell assistant. The current date and time is {timestamp}.
        Solve the tasks that I request you to do.

        You can use the following tools to accomplish the tasks:
        {tools_instructions}
                            
        Rules:
        You can use only the tools provided in this prompt to accomplish the tasks
        If you need to use tools your response must be in JSON format with this structure: {{ "tool": "...", "arg": "...", "thoughts": "..." }}
        Use the shell and your other tools to gather all the information that you need before starting the actual task and also to double check the results if needed before giving the final answer
        After you completed the task output your final answer to the task in {lang_spec} language without including any json
        You can use markdown to format your final answer
        Answer truthfully with the informations you have
        You cannot use tools and complete the task with your final answer in the same message so remember to use the tools that you need first
        """
    ),
}


@chain
def shy_agent_chain(_):
    llm = get_llm()
    if settings.llm.agent_pattern == "function_call":
        llm = llm.bind_tools(tools)
    template = SYS_TEMPLATES[settings.llm.agent_pattern]
    prompt = ChatPromptTemplate.from_messages(
        [
            ("system", template),
            MessagesPlaceholder("few_shot_examples", optional=True),
            MessagesPlaceholder("history"),
            MessagesPlaceholder("tool_history", optional=True),
        ]
    )
    return prompt | llm



from langgraph.graph import StateGraph, START
from shy_sh.agents.shy_agent.nodes.chatbot import chatbot
from shy_sh.agents.shy_agent.nodes.tools_handler import tools_handler
from shy_sh.agents.shy_agent.edges.final_response import final_response_edge
from shy_sh.agents.shy_agent.edges.tool_calls import tool_calls_edge
from shy_sh.models import State

graph_builder = StateGraph(State)


graph_builder.add_node("chatbot", chatbot)
graph_builder.add_node("tools", tools_handler)

graph_builder.add_edge(START, "chatbot")
graph_builder.add_conditional_edges("chatbot", tool_calls_edge)
graph_builder.add_conditional_edges("tools", final_response_edge)

shy_agent_graph = graph_builder.compile()



import re
import json
from time import strftime
from uuid import uuid4
from langchain_core.messages import HumanMessage, AIMessage, ToolMessage
from shy_sh.settings import settings
from shy_sh.utils import detect_shell, detect_os, run_shell
from shy_sh.agents.tools import tools
from shy_sh.models import ToolRequest


def get_graph_inputs(
    history: list,
    examples: list,
    ask_before_execute: bool,
):
    return {
        "history": history,
        "timestamp": strftime("%Y-%m-%d %H:%M %Z"),
        "ask_before_execute": ask_before_execute,
        "lang_spec": settings.language or "",
        "few_shot_examples": examples,
        "tools_instructions": _format_tools(),
    }


def parse_react_tool(message):
    start_idx = message.content.index("{")
    if start_idx < 0:
        raise ValueError("No tool call found")
    end_idx = start_idx + 1
    open_brackets = 1
    while open_brackets > 0 and end_idx < len(message.content):
        if message.content[end_idx] == "{":
            open_brackets += 1
        elif message.content[end_idx] == "}":
            open_brackets -= 1
        end_idx += 1
    maybe_tool = message.content[start_idx:end_idx]
    try:
        return ToolRequest.model_validate_json(maybe_tool)
    except Exception:
        try:
            maybe_tool = message.content[start_idx : message.content.rindex("}") + 1]
            return ToolRequest.model_validate_json(maybe_tool)
        except Exception:
            maybe_tool = re.sub(r"\\(?!\\)", r"\\\\", maybe_tool)
            return ToolRequest.model_validate_json(maybe_tool)


def has_tool_calls(message):
    if settings.llm.agent_pattern == "function_call":
        if bool(getattr(message, "tool_calls", None)):
            return True
    elif settings.llm.agent_pattern == "flow":
        if (message.content or "").count("```") >= 2:
            return True
    elif settings.llm.agent_pattern == "react":
        try:
            parse_react_tool(message)
            return True
        except Exception:
            pass
    return False


def run_flow_few_shot_examples():
    shell = detect_shell()
    os = detect_os()
    pwd_cmd = "echo %cd%" if shell in ["powershell", "cmd"] else "pwd"
    return [
        HumanMessage(
            content=f"You are on {os} system using {shell} as shell. To begin check the current directory and if you are in a git repository."
        ),
        AIMessage(
            content=f"Ok, to check the shell, I will run:\n```{shell}\n{ pwd_cmd}\n```\nthis command will return the current working directory."
        ),
        HumanMessage(content=run_shell(pwd_cmd)),
        AIMessage(
            content=f"Let's see if we are in a git repository\n```{shell}\ngit rev-parse --abbrev-ref HEAD\n```"
        ),
        HumanMessage(content=run_shell("git rev-parse --abbrev-ref HEAD")),
        AIMessage(content="Well done! 👍\nDo you need anything else?"),
    ]


def run_few_shot_examples():
    if settings.llm.agent_pattern == "flow":
        return run_flow_few_shot_examples()
    shell = detect_shell()
    os = detect_os()
    actions = [
        {
            "tool": "shell",
            "arg": "echo test" if shell in ["powershell", "cmd"] else "echo $SHELL",
            "thoughts": "I'm checking the current shell",
        },
        {
            "tool": "shell",
            "arg": "echo %cd%" if shell in ["powershell", "cmd"] else "pwd",
            "thoughts": "I'm checking the current working directory",
        },
        {
            "tool": "shell",
            "arg": "git rev-parse --abbrev-ref HEAD",
            "thoughts": "I'm checking if it's a git repository",
        },
    ]
    result = []
    result.append(
        HumanMessage(
            content=f"You are on {os} system using {shell} as shell. Check your tools to get started."
        )
    )
    for action in actions:
        uid = str(uuid4())
        ai_message, response = _run_example(action, uid)
        result.append(ai_message)
        if settings.llm.agent_pattern == "react":
            result.append(HumanMessage(content=f"Tool response:\n{response}"))
        elif settings.llm.agent_pattern == "function_call":
            result.append(ToolMessage(content=response, tool_call_id=uid))
    result.append(AIMessage(content="All set! 👍"))
    return result


def _run_example(action, uid):
    ai_message = AIMessage(
        content="",
        tool_calls=[
            {
                "id": uid,
                "type": "tool_call",
                "name": action["tool"],
                "args": {"arg": action["arg"]},
            }
        ],
    )
    if settings.llm.agent_pattern == "react":
        ai_message.content = json.dumps(action)
        ai_message.tool_calls = []
    return ai_message, run_shell(action["arg"])


def _format_tools():
    if settings.llm.agent_pattern == "function_call":
        return None
    return "\n".join(map(lambda tool: f'- "{tool.name}": {tool.description}', tools))



from rich import print
from langgraph.graph import END
from shy_sh.models import State, ToolMeta
from shy_sh.settings import settings
from shy_sh.utils import syntax


def final_response_edge(state: State):
    last_tool = state["tool_history"][-1]
    artifact = getattr(last_tool, "artifact", None)
    if isinstance(artifact, ToolMeta) and artifact.stop_execution:
        message = last_tool.content
        if settings.llm.agent_pattern == "react":
            message = message.replace("Tool response:\n", "", 1)
        if not artifact.skip_print:
            print(syntax(f"🤖: {message}"))
        return END
    print()
    return "chatbot"



from langchain_core.runnables import chain
from langchain_core.output_parsers import StrOutputParser
from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from shy_sh.agents.llms import get_llm
from textwrap import dedent


msg_template = dedent(
    """
    Output only a block of code like this:
    ```sh
    #!/bin/sh
    set -e
    [your shell script]
    ```

    This is a template for sh, but you should use the right shell syntaxt depending on your system.
    Don't install new packages if not explicitly requested.
    Write a shell script that accomplishes the task.

    Task: {input}
    """
)


@chain
def shexpert_chain(_):
    llm = get_llm()
    prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "You are a shell expert. The current date and time is {timestamp}\nYou are running on {system} using {shell} as shell",
            ),
            MessagesPlaceholder("history"),
            ("human", msg_template),
        ]
    )
    return prompt | llm | StrOutputParser()



import os
import sys
import pyperclip
from typing import Annotated
from tempfile import NamedTemporaryFile
from rich import print
from rich.live import Live
from langgraph.prebuilt import InjectedState
from langchain.tools import tool
from shy_sh.models import State, ToolMeta
from shy_sh.utils import ask_confirm, tools_to_human, syntax, run_python, parse_code
from shy_sh.agents.chains.python_expert import pyexpert_chain
from shy_sh.agents.chains.explain import explain


@tool(response_format="content_and_artifact")
def python_expert(arg: str, state: Annotated[State, InjectedState]):
    """to delegate the task to a python expert that can write and execute python code, use only if you cant resolve the task with shell, just explain what do you want to achieve in a short sentence in the arg without including any python code"""
    print(f"🐍 [bold yellow]Generating python script...[/bold yellow]\n")
    inputs = {
        "timestamp": state["timestamp"],
        "history": tools_to_human(state["history"] + state["tool_history"]),
        "input": arg,
    }
    code = ""
    with Live(vertical_overflow="visible") as live:
        for chunk in pyexpert_chain.stream(inputs):
            code += chunk  # type: ignore
            live.update(syntax(code, "python"))
        code = parse_code(code)
        live.update(syntax(code.strip(), "python"))

    confirm = "y"
    if state["ask_before_execute"]:
        confirm = ask_confirm()
    print()
    if confirm == "n":
        return "Command canceled by user", ToolMeta(
            stop_execution=True, skip_print=True
        )
    elif confirm == "c":
        pyperclip.copy(code)
        return "Script copied to the clipboard!", ToolMeta(stop_execution=True)
    elif confirm == "e":
        inputs = {
            "task": arg,
            "script_type": "python script",
            "script": code,
            "timestamp": state["timestamp"],
        }
        ret = explain(inputs)
        if ret:
            return ret

    if sys.version_info >= (3, 12):
        with NamedTemporaryFile("w+", suffix=".py", delete_on_close=False) as file:
            file.write(code)
            file.close()
            os.chmod(file.name, 0o755)
            result = run_python(file.name)

            if len(result) > 20000:
                print("\n🐳 [bold red]Output too long! It will be truncated[/bold red]")
                result = (
                    result[:9000]
                    + "\n...(OUTPUT TOO LONG TRUNCATED!)...\n"
                    + result[-9000:]
                )

    else:
        with NamedTemporaryFile("w+", suffix=".py", delete=False) as file:
            file.write(code)
            file.close()
            os.chmod(file.name, 0o755)
            result = run_python(file.name)

            if len(result) > 20000:
                print("\n🐳 [bold red]Output too long! It will be truncated[/bold red]")
                result = (
                    result[:9000]
                    + "\n...(OUTPUT TOO LONG TRUNCATED!)...\n"
                    + result[-9000:]
                )
            os.unlink(file.name)

    ret = f"\nScript executed:\n```python\n{code.strip()}\n```\n\nOutput:\n{result}"
    if len(ret) > 20000:
        ret = f"Output:\n{result}"
    return (
        ret,
        ToolMeta(
            executed_scripts=[{"script": code, "type": "python", "result": result}]
        ),
    )
